
// a simple C-pthreads program using fork-join synchronization

// note: only reason we are choosing the thread-count statically is to
//       stack-allocate the args and threads arrays -- we could
//       easily heap-allocate them or use a MAX_THREADS value or something

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#define NUM_THREADS 4

#ifdef SHOW_RESULTS
#define ARRAY_SIZE 160
void finished_iter(int th, int ans, int iter) {
  if(iter % 10 != 0)
    return;
  printf("thread: %d, sum so far: %d\n",th, ans);
  fflush(stdout);  
}
#else
#define ARRAY_SIZE 16000
void finished_iter(int th, int ans, int iter) {}
#endif

#if (ARRAY_SIZE % NUM_THREADS)
#error "will miss some array elements"
#else
#define PER_THREAD_SIZE (ARRAY_SIZE / NUM_THREADS)
#endif

struct SlowSumArg {
  int * arr;
  int len;
  int which_thread;
  int ans;
};
typedef struct SlowSumArg * slowsumarg_t;

void* slow_sum(void * a) {
  slowsumarg_t args = (slowsumarg_t)a;
  int ans = 0;
  int i;
  for(i=0; i<args->len; ++i) {
    int num = args->arr[i];
    while(num) { // a silly, slow algorithm for ans += num
      --num;
      ++ans;
    }
    finished_iter(args->which_thread,ans,i);
  }
  args->ans = ans;
  return NULL; 
}

void big_sum() {
  int *             arr = (int*)malloc(ARRAY_SIZE*sizeof(int));
  int               i;
  struct SlowSumArg args[NUM_THREADS];
  pthread_t         threads[NUM_THREADS];
  int               total_ans = 0;
  // sequential initialization (could be a bottleneck)
  for(i=0; i < ARRAY_SIZE; ++i)
    arr[i] = (i+1)*50;
  for(i=0; i < NUM_THREADS; ++i) {
    args[i].arr = arr+i*PER_THREAD_SIZE;
    args[i].len = PER_THREAD_SIZE;
    args[i].which_thread = i;
    args[i].ans = 0;
    if(pthread_create(&threads[i],NULL,slow_sum,&args[i])) {
      fprintf(stderr,"could not create a thread");
      exit(1);
    }
  }
  for(i=0; i < NUM_THREADS; ++i)
    pthread_join(threads[i],NULL); 
  for(i=0; i < NUM_THREADS; ++i)
    total_ans += args[i].ans;
  printf("total answer: %d\n",total_ans);
}

int main(int argc, char**argv) {
  big_sum();
  return 0;
}
